(* $Id: toploop.ml,v 1.7 2016/06/14 07:26:12 garrigue Exp $ *)

open Terms
open Parser
open Eval
open Printer

exception Next of (string * command * string) Env.t
let next procs = raise (Next procs)

let rec not_seq = function
    Seq  _ -> false
  | _ -> true

let tracep = ref false

let trace ~env c =
  if !tracep && not_seq c then begin
    Format.printf "+ @[%a@]@." print_cmd c;
    let rec loop () =
      Printf.printf "? "; flush stdout;
      let q = input_line stdin in
      try match lexer(Stream.of_string q) with parser
	[< 'Atm"q" >] -> raise Sys.Break
      | [< 'Ident x >] ->
	  Format.printf "= %a@." print_val (get ~env x);
	  raise Exit
      | [< 'EOF >] -> ()
      with Exit | Stream.Failure | Stream.Error _ -> loop ()
    in loop ()
  end

let () = trace_fun := trace

let cstrm = ref [< >]

let rec toploop ?strm procs =
  try
    let s = 
      match strm with Some s -> s
      | None ->
          Printf.printf "> "; flush stdout;
          let cmd = input_line stdin in
          if cmd = "" then  next procs else begin
	    cstrm := Stream.of_string cmd;
            lexer !cstrm
	  end
    in
    match s with parser
    | [< 'Kwd"define"; 'Atm f >] ->
        let s =
          if strm <> None then s else begin
	    cstrm := Stream.of_channel stdin;
	    lexer !cstrm
	  end
	in
        let p = program s in
        Format.printf "= %s defined.@." f;
        next (Env.add f p procs)
    | [< 'Atm"quit" >] ->
        procs
    | [< 'Atm"load"; 'Atm file >] ->
        let file =
          if Sys.file_exists file then file else
          let fw = file ^ ".whl" in
          if Sys.file_exists fw then fw else
          let ft = file ^ ".txt" in
          if Sys.file_exists ft then ft else
          failwith ("File not found: " ^ file)
        in
        let ic = open_in file in
	cstrm := Stream.of_channel ic;
        let strm = lexer !cstrm in
        let procs = toploop ~strm procs in
        close_in ic;
        next procs
    | [< 'Atm"trace" >] ->
	tracep := not !tracep;
	Format.printf "Trace is %s.@." (if !tracep then "on" else "off");
	next procs
    | [< 'Atm f; e = expr >] ->
        let r = eval ~procs f (eval_expr ~env:Env.empty e) in
        Format.printf "= %a@." print_val r;
        if strm = None && Stream.next s <> EOF then
          prerr_endline "Garbage at end of line.";
        next procs
    | [< e = expr >] ->
        let r = eval_expr ~env:Env.empty e in
        Format.printf "= %a@." print_val r;
        if strm = None && Stream.next s <> EOF then
          prerr_endline "Garbage at end of line.";
        next procs
    | [< 'EOF >] -> procs
  with
    End_of_file ->
      if strm = None then prerr_endline "Bye.";
      procs
  | Next procs ->
      toploop ?strm procs
  | Stream.Failure | Stream.Error "" ->
      let (n,c) = get_pos !cstrm in
      Printf.eprintf "Parsing error before line %d, char %d.\n" n c;
      flush stderr;
      if strm = None then toploop procs else procs
  | Stream.Error s ->
      let (n,c) = get_pos !cstrm in
      Printf.eprintf "Parsing error before line %d, char %d: %s.\n" n c s;
      flush stderr;
      if strm = None then toploop procs else procs
  | exn ->
      prerr_endline (Printexc.to_string exn);
      if strm = None then toploop procs else procs

let start () =
  Sys.catch_break true;
  toploop Env.empty

let _ = start ()
